#ifndef _KERNEL_FS_H
#define _KERNEL_FS_H 1

#include <stdint.h>
#include <sys/types.h>
#include <kernel/memory.h>
#include <sys/os16.h>
#include <sys/stat.h>
#include <stdint.h>
#include <const.h>
#include <stdio.h>
#include <limits.h>

//----------------------------------------------------------------------
#define SB_MAX_INODE_BLOCKS          1  // 8192 inodes max.
#define SB_MAX_ZONE_BLOCKS           1  // 8192 data-zones max.
#define SB_BLOCK_SIZE             1024  // Fixed for Minix file system.
#define SB_MAX_ZONE_SIZE          2048  // log2 max is 1.
#define SB_MAP_INODE_SIZE         (SB_MAX_INODE_BLOCKS*512) // [1]
#define SB_MAP_ZONE_SIZE          (SB_MAX_ZONE_BLOCKS*512)  // [1]
//
// [1] blocks * (1024 * 8 / 16) = number of bits, divided 16.
//
//----------------------------------------------------------------------
#define INODE_MAX_INDIRECT_ZONES  (SB_MAX_ZONE_SIZE/2)      // [2]

#define INODE_MAX_REFERENCES      0xFF
//
// [2] number of zone pointers contained inside a zone, used
//     as an indirect inode list (a pointer = 16 bits = 2 bytes).
//
//----------------------------------------------------------------------
typedef uint16_t        zno_t;          // Zone number.
//----------------------------------------------------------------------
// The structured type `inode_t' must be pre-declared here, because
// the type sb_t, described before the inode structure, has a member
// pointing to a type `inode_t'. So, must be declared previously
// the type `inode_t' as made of a type `struct inode', then the
// structure `inode' can be described. But for a matter of coherence,
// all other structured data declared inside this file follow the
// same procedure.
//
typedef struct sb        sb_t;
typedef struct inode     inode_t;
typedef struct file      file_t;
typedef struct fd        fd_t;
typedef struct directory directory_t;
//----------------------------------------------------------------------
#define SB_MAX_SLOTS  2                 // Handle max 2 file systems.

struct sb {                             // File system super block:
    uint16_t      inodes;               // inodes available;
    uint16_t      zones;                // zones available (disk size);
    uint16_t      map_inode_blocks;     // inode bit map blocks;
    uint16_t      map_zone_blocks;      // data-zone bit map blocks;
    uint16_t      first_data_zone;      // first data-zone;
    uint16_t      log2_size_zone;       // log_2 (size_zone/block_size);
    uint32_t      max_file_size;        // max file size in bytes;
    uint16_t      magic_number;         // file system magic number.
    //------------------------------------------------------------------
    // Extra management data, not saved inside the file system
    // super block.
    //------------------------------------------------------------------
    dev_t         device;                       // FS device [3]
    inode_t      *inode_mounted_on;             // [4]
    blksize_t     blksize;                      // Calculated zone size.
    int           options;                      // [5]
    uint16_t      map_inode[SB_MAP_INODE_SIZE];
    uint16_t      map_zone[SB_MAP_ZONE_SIZE];
    char          changed;
};

extern sb_t    sb_table[SB_MAX_SLOTS];
//
// [3] the member `device' must be kept at the same position, because
//     it is used to calculate the super block header size, saved on
//     disk.
//
// [4] If this pointer is not NULL, the super block is related to a
//     device mounted on a directory. The inode of such directory is
//     recorded here. Please note that it is type `void *', instead of
//     type `inode_t', because type `inode_t' is declared after type
//     `sb_t'.
//     Please note that the type `sb_t' is declared before the
//     type `inode_t', but this member points to a type `inode_t'.
//     This is the reason because it was necessary to declare first
//     the type `inode_t' as made of `struct inode', to be described
//     later. For coherence, all derived type made of structured data,
//     are first declared as structure, and then, later, described.
//
// [5] Mount options can be only `MOUNT_DEFAULT' or `MOUNT_RO',
//     as defined inside file `lib/sys/os16.h'.
//
//----------------------------------------------------------------------
#define INODE_MAX_SLOTS           (8 * OPEN_MAX)

struct inode {                          // Inode (32 byte total):
    mode_t        mode;                 // file type and permissions;
    uid_t         uid;                  // user ID (16 bit);
    ssize_t       size;                 // file size in bytes;
    time_t        time;                 // file data modification time;
    uint8_t       gid;                  // group ID (8 bit);
    uint8_t       links;                // links to the inode;
    zno_t         direct[7];            // direct zones;
    zno_t         indirect1;            // indirect zones;
    zno_t         indirect2;            // double indirect zones.
    //------------------------------------------------------------------
    // Extra management data, not saved inside the disk file system.
    //------------------------------------------------------------------
    sb_t         *sb;                   // Inode's super block. [7]
    ino_t         ino;                  // Inode number.
    sb_t         *sb_attached;          // [8]
    blkcnt_t      blkcnt;               // Rounded size/blksize.
    unsigned char references;           // Run time active references.
    char          changed;              // 1 == to be saved.
};

extern inode_t inode_table[INODE_MAX_SLOTS];
//
// [7] the member `sb' must be kept at the same position, because
//     it is used to calculate the inode header size, saved on disk.
//
// [8] If the inode is a mount point for another device, the other
//     super block pointer is saved inside `sb_attached'.
//
//----------------------------------------------------------------------
#define FILE_MAX_SLOTS             (16 * OPEN_MAX)

struct file {
    int           references;
    off_t         offset;               // File position.
    int           oflags;               // Open mode: r/w/r+w [9]
    inode_t      *inode;
};

extern file_t file_table[FILE_MAX_SLOTS];
//
// [9] the member `oflags' can get only O_RDONLY, O_WRONLY, O_RDWR,
//     (from header `fcntl.h') combined with OR binary operator.
//
//----------------------------------------------------------------------
struct fd {
    int           fl_flags;             // File status flags and file
                                        // access modes. [10]
    int           fd_flags;             // File descriptor flags:
                                        // currently only FD_CLOEXEC.
    file_t       *file;                 // Pointer to the file table.
};
//
// [10] the member `fl_flags' can get only O_RDONLY, O_WRONLY, O_RDWR,
//      O_CREAT, O_EXCL, O_NOCTTY, O_TRUNC and O_APPEND
//      (from header `fcntl.h') combined with OR binary
//      operator. Options like O_DSYNC, O_NONBLOCK, O_RSYNC and O_SYNC
//      are not taken into consideration by os16.
//
//----------------------------------------------------------------------
struct directory {                      // Directory entry:
    ino_t         ino;                  // inode number;
    char          name[NAME_MAX];       // file name.
};
//----------------------------------------------------------------------
sb_t    *sb_reference    (dev_t device);
sb_t    *sb_mount        (dev_t device, inode_t **inode_mnt,
                          int options);
sb_t    *sb_get          (dev_t device, sb_t *sb);
int      sb_save         (sb_t *sb);
int      sb_zone_status  (sb_t *sb, zno_t zone);
int      sb_inode_status (sb_t *sb, ino_t ino);
//----------------------------------------------------------------------
zno_t    zone_alloc           (sb_t *sb);
int      zone_free            (sb_t *sb, zno_t zone);
int      zone_read            (sb_t *sb, zno_t zone, void *buffer);
int      zone_write           (sb_t *sb, zno_t zone, void *buffer);
//----------------------------------------------------------------------
inode_t *inode_reference      (dev_t device, ino_t ino);
inode_t *inode_alloc          (dev_t device, mode_t mode, uid_t uid);
int      inode_free           (inode_t *inode);
inode_t *inode_get            (dev_t device, ino_t ino);
int      inode_save           (inode_t *inode);
int      inode_put            (inode_t *inode);
int      inode_truncate       (inode_t *inode);
zno_t    inode_zone           (inode_t *inode, zno_t fzone, int write);
inode_t *inode_stdio_dev_make (dev_t device, mode_t mode);
blkcnt_t inode_fzones_read    (inode_t *inode, zno_t zone_start,
                               void *buffer, blkcnt_t blkcnt);
blkcnt_t inode_fzones_write   (inode_t *inode, zno_t zone_start,
                               void *buffer, blkcnt_t blkcnt);
ssize_t  inode_file_read      (inode_t *inode, off_t offset,
                               void *buffer, size_t count, int *eof);
ssize_t  inode_file_write     (inode_t *inode, off_t offset,
                               void *buffer, size_t count);
int      inode_check          (inode_t *inode, mode_t type,
                               int perm, uid_t uid);
int      inode_dir_empty      (inode_t *inode);
//----------------------------------------------------------------------
file_t  *file_reference       (int fno);
file_t  *file_stdio_dev_make  (dev_t device, mode_t mode, int oflags);
//----------------------------------------------------------------------
inode_t *path_inode      (pid_t pid, const char *path);
int      path_chdir      (pid_t pid, const char *path);
dev_t    path_device     (pid_t pid, const char *path);
int      path_full       (const char *path,
                          const char *path_cwd,
                          char *full_path);
int      path_fix        (char *path);
inode_t *path_inode_link (pid_t pid, const char *path, inode_t *inode,
                          mode_t mode);
int      path_link       (pid_t pid, const char *path_old,
                          const char *path_new);
int      path_mkdir      (pid_t pid, const char *path, mode_t mode);
int      path_mknod      (pid_t pid, const char *path, mode_t mode,
                          dev_t device);
int      path_mount      (pid_t pid, const char *path_dev,
                          const char *path_mnt,
                          int options);
int      path_umount     (pid_t pid, const char *path_mnt);
int      path_stat       (pid_t pid, const char *path,
                          struct stat *buffer);
int      path_chmod      (pid_t pid, const char *path, mode_t mode);
int      path_chown      (pid_t pid, const char *path, uid_t uid,
                          gid_t gid);
int      path_unlink     (pid_t pid, const char *path);
//----------------------------------------------------------------------
fd_t    *fd_reference (pid_t pid, int *fdn);
int      fd_chmod     (pid_t pid, int fdn, mode_t mode);
int      fd_chown     (pid_t pid, int fdn, uid_t uid, gid_t gid);
int      fd_close     (pid_t pid, int fdn);
int      fd_fcntl     (pid_t pid, int fdn, int cmd, int arg);
int      fd_dup       (pid_t pid, int fdn_old, int fdn_min);
int      fd_dup2      (pid_t pid, int fdn_old, int fdn_new);
off_t    fd_lseek     (pid_t pid, int fdn, off_t offset, int whence);
int      fd_open      (pid_t pid, const char *path, int oflags,
                       mode_t mode);
ssize_t  fd_read      (pid_t pid, int fdn, void *buffer, size_t count,
                       int *eof);
int      fd_stat      (pid_t pid, int fdn, struct stat *buffer);
ssize_t  fd_write     (pid_t pid, int fdn, const void *buffer,
                       size_t count);
//----------------------------------------------------------------------

#endif
