#include <kernel/proc.h>
#include <kernel/k_libc.h>
#include <errno.h>
#include <fcntl.h>
//----------------------------------------------------------------------
int
fd_open (pid_t pid, const char *path, int oflags, mode_t mode)
{
    proc_t   *ps;
    inode_t  *inode;
    int       status;
    file_t   *file;
    fd_t     *fd;
    int       fdn;
    char      full_path[PATH_MAX];
    int       perm;
    tty_t    *tty;
    mode_t    umask;
    int       errno_save;
    //
    // Get process.
    //
    ps = proc_reference (pid);
    //
    // Correct the mode with the umask. As it is not a directory, to the
    // mode are removed execution and sticky permissions.
    //
    umask = ps->umask | 01111;
    mode &= ~umask;
    //
    // Check open options.
    //
    if (oflags & O_WRONLY)
      {
        //
        // The file is to be opened for write, or for read/write.
        // Try to get inode.
        //
        inode = path_inode (pid, path);
        if (inode == NULL)
          {
            //
            // Cannot get the inode. See if there is the creation
            // option.
            //
            if (oflags & O_CREAT)
              {
                //
                // Try to create the missing inode: the file must be a
                // regular one, so add the mode.
                //
                path_full (path, ps->path_cwd, full_path);
                inode = path_inode_link (pid, full_path, NULL,
                                         (mode | S_IFREG));
                if (inode == NULL)
                  {
                    //
                    // Sorry: cannot create the inode! Variable `errno'
                    // is already set by `path_inode_link()'.
                    //
                    errset (errno);
                    return (-1);
                  }
              }
            else
              {
                //
                // Cannot open the inode. Variable `errno'
                // should be already set by `path_inode()'.
                //
                errset (errno);
                return (-1);
              }
          }
        //
        // The inode was read or created: check if it must be
        // truncated. It can be truncated only if it is a regular
        // file.
        //
        if (oflags & O_TRUNC && inode->mode & S_IFREG)
          {
            //
            // Truncate inode.
            //
            status = inode_truncate (inode);
            if (status != 0)
              {
                //
                // Cannot truncate the inode: release it and return.
                // But this error should never happen, because the
                // function `inode_truncate()' will not return any
                // other value than zero.
                //
                errno_save = errno;
                inode_put (inode);
                errset (errno_save);
                return (-1);
              }
          }
      }
    else
      {
        //
        // The file is to be opened for read, but not for write.
        // Try to get inode.
        //
        inode = path_inode (pid, path);
        if (inode == NULL)
          {
            //
            // Cannot open the file.
            //
            errset (errno);
            return (-1);
          }
      }
    //
    // An inode was opened: check type and access permissions.
    // All file types are good, even directories, as the type
    // DIR is implemented through file descriptors.
    //
    perm = 0;
    if (oflags & O_RDONLY) perm |= 4;
    if (oflags & O_WRONLY) perm |= 2;
    status = inode_check (inode, S_IFMT, perm, ps->uid);
    if (status != 0)
      {
        //
        // The file type is not correct or the user does not have
        // permissions.
        //
        return (-1);
      }
    //
    // Allocate the file, inside the file table.
    //
    file = file_reference (-1);
    if (file == NULL)
      {
        //
        // Cannot allocate the file inside the file table: release the
        // inode, update `errno' and return.
        //
        inode_put (inode);
        errset (ENFILE);        // Too many files open in system.
        return (-1);
      }
    //
    // Put some data inside the file item. Only options
    // O_RDONLY and O_WRONLY are kept here, because the O_APPEND
    // is saved inside the file descriptor table.
    //
    file->references = 1;
    file->oflags     = (oflags & (O_RDONLY | O_WRONLY));
    file->inode      = inode;
    //
    // Allocate the file descriptor: variable `fdn' will be modified
    // by the call to `fd_reference()'.
    //
    fdn = -1;
    fd = fd_reference (pid, &fdn);
    if (fd == NULL)
      {
        //
        // Cannot allocate the file descriptor: remove the item from
        // file table.
        //
        file->references = 0;
        file->oflags     = 0;
        file->inode      = NULL;
        //
        // Release the inode.
        //
        inode_put (inode);
        //
        // Return an error.
        //
        errset (EMFILE);        // Too many open files.
        return (-1);
      }
    //
    // File descriptor allocated: put some data inside the
    // file descriptor item.
    //
    fd->fl_flags = (oflags & (O_RDONLY | O_WRONLY | O_APPEND));
    fd->fd_flags = 0;
    fd->file = file;
    fd->file->offset = 0;
    //
    // Check if it is a terminal (currently only consoles), if it is
    // opened for read and write, and if it have to be set as the
    // controlling terminal. This thing is done here because there is
    // not a real device driver.
    //
    if ((S_ISCHR (inode->mode)) &&
        (oflags & O_RDONLY)     &&
        (oflags & O_WRONLY))
      {
        //
        // The inode is a character special file (related to a character
        // device), opened for read and write!
        //
        if ((inode->direct[0] & 0xFF00) == (DEV_CONSOLE_MAJOR << 8))
          {
            //
            // It is a terminal (currently only consoles are possible).
            // Get the tty reference.
            //
            tty = tty_reference ((dev_t) inode->direct[0]);
            //
            // Verify that the terminal is not already the controlling
            // terminal of some process group.
            //
            if (tty->pgrp == 0)
              {
                //
                // The terminal is free: verify if the current process
                // needs a controlling terminal.
                //
                if (ps->device_tty == 0 && ps->pgrp == pid)
                  {
                    //
                    // It is a group leader with no controlling
                    // terminal: set the controlling terminal.
                    //
                    ps->device_tty = inode->direct[0];
                    tty->pgrp      = ps->pgrp;
                  }
              }
          }
      }
    //
    // Return the file descriptor.
    //
    return (fdn);
}
