#include <kernel/fs.h>
#include <errno.h>
#include <kernel/devices.h>
//----------------------------------------------------------------------
sb_t *
sb_mount (dev_t device, inode_t **inode_mnt, int options)
{
    sb_t   *sb;
    ssize_t size_read;
    addr_t  start;
    int     m;
    size_t  size_sb;
    size_t  size_map;
    //
    // Find if it is already mounted.
    //
    sb = sb_reference (device);
    if (sb != NULL)
      {
        errset (EBUSY);         // Device or resource busy: device
        return (NULL);          // already mounted.
      }
    //
    // Find if `*inode_mnt' is already mounting something.
    //
    if (*inode_mnt != NULL && (*inode_mnt)->sb_attached != NULL)
      {
        errset (EBUSY);         // Device or resource busy: mount point
        return (NULL);          // already used.
      }
    //
    // The inode is not yet mounting anything, or it is new: find a free
    // slot inside the super block table.
    //
    sb = sb_reference ((dev_t) -1);
    if (sb == NULL)
      {
        errset (EBUSY);         // Device or resource busy:
        return (NULL);          // no free slots.
      }
    //
    // A free slot was found: the super block header must be loaded, but
    // before it is necessary to calculate the header size to be read.
    //
    size_sb = offsetof (sb_t, device);
    //
    // Then fix the starting point.
    //
    start  = 1024;                      // After boot block.
    //
    // Read the file system super block header.
    //
    size_read = dev_io ((pid_t) -1, device, DEV_READ, start, sb,
                        size_sb, NULL);
    if (size_read != size_sb)
      {
        errset (EIO);                   // I/O error.
        return (NULL);
      }
    //
    // Save some more data.
    //
    sb->device           =  device;
    sb->options          =  options;
    sb->inode_mounted_on = *inode_mnt;
    sb->blksize          =  (1024 << sb->log2_size_zone);
    //
    // Check if the super block data is valid.
    //
    if (sb->magic_number != 0x137F)
      {
        errset (ENODEV);        // No such device: unsupported
        sb->device = 0;         // file system type.
        return (NULL);
      }
    if (sb->map_inode_blocks > SB_MAX_INODE_BLOCKS)
      {
        errset (E_MAP_INODE_TOO_BIG);
        return (NULL);
      }
    if (sb->map_zone_blocks > SB_MAX_ZONE_BLOCKS)
      {
        errset (E_MAP_ZONE_TOO_BIG);
        return (NULL);
      }
    if (sb->blksize > SB_MAX_ZONE_SIZE)
      {
        errset (E_DATA_ZONE_TOO_BIG);
        return (NULL);
      }
    //
    // A right super block header was loaded from disk, now load the
    // super block inode bit map.
    //
    start  = 1024;                              // After boot block.
    start += 1024;                              // After super block.
    for (m = 0; m < SB_MAP_INODE_SIZE; m++)     //
      {                                         // Reset map in memory,
        sb->map_inode[m] = 0xFFFF;              // before loading.
      }                                         //
    size_map = sb->map_inode_blocks * 1024;
    size_read = dev_io ((pid_t) -1, sb->device, DEV_READ, start,
                        sb->map_inode, size_map, NULL);
    if (size_read != size_map)
      {
        errset (EIO);                   // I/O error.
        return (NULL);
      }
    //
    // Load the super block zone bit map.
    //
    start  = 1024;                              // After boot block.
    start += 1024;                              // After super block.
    start += (sb->map_inode_blocks * 1024);     // After inode bit map.
    for (m = 0; m < SB_MAP_ZONE_SIZE; m++)      //
      {                                         // Reset map in memory,
        sb->map_zone[m] = 0xFFFF;               // before loading.
      }                                         //
    size_map = sb->map_zone_blocks * 1024;
    size_read = dev_io ((pid_t) -1, sb->device, DEV_READ, start,
                        sb->map_zone, size_map, NULL);
    if (size_read != size_map)
      {
        errset (EIO);                   // I/O error.
        return (NULL);
      }
    //
    // Check the inode that should mount the super block. If
    // `*inode_mnt' is `NULL', then it is meant to be the first mount of
    // the root file system. In such case, the inode must be loaded too,
    // and the value for `*inode_mnt' must be modified.
    //
    if (*inode_mnt == NULL)
      {
        *inode_mnt = inode_get (device, 1);
      }
    //
    // Check for a valid value.
    //
    if (*inode_mnt == NULL)
      {
        //
        // This is bad!
        //
        errset (EUNKNOWN);              // Unknown error.
        return (NULL);
      }
    //
    // A valid inode is available for the mount.
    //
    (*inode_mnt)->sb_attached  = sb;
    //
    // Return the super block pointer.
    //
    return (sb);
}
