#include <kernel/proc.h>
#include <kernel/k_libc.h>
#include <stdint.h>
//----------------------------------------------------------------------
extern uint16_t _ksp;
//----------------------------------------------------------------------
void
proc_scheduler (uint16_t *sp, segment_t *segment_d)
{
    //
    // The process is identified from the data and stack segment.
    //
    pid_t prev;
    pid_t next;
    //
    static unsigned long int previous_clock;
           unsigned long int current_clock;
    //
    // Check if current data segments are right.
    //
    if (es () != ds () || ss () != ds ())
      {
        k_printf ("\n");
        k_printf ("Kernel panic: ES, DS, SS are different!\n");
        k_exit (0);
      }
    //
    // Search the data segment inside the process table.
    // Must be done here, because the subsequent call to
    // proc_sch_signals() will remove the segment numbers
    // from a zombie process.
    //
    prev = proc_find (*segment_d);
    //
    // Take care of sleeping processes: wake up if sleeping time
    // elapsed.
    //
    proc_sch_timers ();
    //
    // Take care of pending signals.
    //
    proc_sch_signals ();
    //
    // Take care input from terminals.
    //
    proc_sch_terminals ();
    //
    // Update the CPU time usage.
    //
    current_clock   = k_clock ();
    proc_table[prev].usage += current_clock - previous_clock;
    previous_clock  = current_clock;
    //
    // Scan for a next process.
    //
    for (next = prev+1; next != prev; next++)
      {
        if (next >= PROCESS_MAX)
          {
            next = -1; // At the next loop, `next' will be zero.
            continue;
          }
        if      (proc_table[next].status == PROC_EMPTY)
          {
            continue;
          }
        else if (proc_table[next].status == PROC_CREATED) 
          {
            continue;
          }
        else if (proc_table[next].status == PROC_READY) 
          {
            if (proc_table[prev].status == PROC_RUNNING)
              {
                proc_table[prev].status = PROC_READY;
              }
            proc_table[prev].sp     = *sp;
            proc_table[next].status = PROC_RUNNING;
            proc_table[next].ret    = 0;
            *segment_d      = proc_table[next].segment_d;
            *sp             = proc_table[next].sp;
            break;
          }
        else if (proc_table[next].status == PROC_RUNNING)
          {
            if (proc_table[prev].status == PROC_RUNNING)
              {
                k_printf ("Kernel alert: process %i ", prev);
                k_printf ("and %i \"running\"!\r", next);
                proc_table[prev].status = PROC_READY;
              }
            proc_table[prev].sp     = *sp;
            proc_table[next].status = PROC_RUNNING;
            proc_table[next].ret    = 0;
            *segment_d      = proc_table[next].segment_d;
            *sp             = proc_table[next].sp;
            break;
          }
        else if (proc_table[next].status == PROC_SLEEPING)
          {
            continue;
          }
        else if (proc_table[next].status == PROC_ZOMBIE)
          {
            continue;
          }
      }
    //
    // Check again if the next process is set to running, otherwise set
    // the kernel to such value!
    //
    next = proc_find (*segment_d);
    if (proc_table[next].status != PROC_RUNNING)
      {
        proc_table[0].status = PROC_RUNNING;
        *segment_d   = proc_table[0].segment_d;
        *sp          = proc_table[0].sp;
      }
    //
    // Save kernel stack pointer.
    //
    _ksp = proc_table[0].sp;
    //
    // At the end, must inform the PIC 1, with message «EOI».
    //
    out_8 (0x20, 0x20);
}
