#include <kernel/proc.h>
#include <kernel/k_libc.h>
//----------------------------------------------------------------------
void
proc_sys_exit (pid_t pid, int status)
{
  pid_t  child;
  pid_t  parent = proc_table[pid].ppid;
  pid_t  extra;
  int    proc_count;
  int    sigchld = 0;
  int    fdn;
  tty_t *tty;
  //
  proc_table[pid].status     = PROC_ZOMBIE;
  proc_table[pid].ret        = status;
  proc_table[pid].sig_status =  0;
  proc_table[pid].sig_ignore =  0;
  //
  // Close files.
  //
  for (fdn = 0; fdn < OPEN_MAX; fdn++)
    {
      fd_close (pid, fdn);
    }
  //
  // Close current directory.
  //
  inode_put (proc_table[pid].inode_cwd);
  //
  // Close the controlling terminal, if it is a process leader with
  // such a terminal.
  //
  if (proc_table[pid].pgrp == pid && proc_table[pid].device_tty != 0)
    {
      tty = tty_reference (proc_table[pid].device_tty);
      //
      // Verify.
      //
      if (tty == NULL)
        {
          //
          // Show a kernel message.
          //
          k_printf ("kernel alert: cannot find the terminal item "
                    "for device 0x%04x!\n",
                    (int) proc_table[pid].device_tty);
        }
      else if (tty->pgrp != pid)
        {
          //
          // Show a kernel message.
          //
          k_printf ("kernel alert: terminal device 0x%04x should "
                    "be associated to the process group %i, but it "
                    "is instead related to process group %i!\n",
                    (int) proc_table[pid].device_tty, (int) pid,
                    (int) tty->pgrp);
        }
      else
        {
          tty->pgrp = 0;
        }
    }
  //
  // Free data segment memory.
  //
  mb_free (proc_table[pid].address_d, proc_table[pid].size_d);
  //
  // Free code segment memory if it is
  // different from the data segment.
  //
  if (proc_table[pid].segment_i != proc_table[pid].segment_d)
    {
      //
      // Must verify if no other process is using the same memory.
      // The proc_count variable is incremented for processes
      // active, ready or sleeping: the current process is already
      // set as zombie, and is not counted.
      //
      for (proc_count = 0, extra = 0; extra < PROCESS_MAX; extra++)
        {
          if (proc_table[extra].status == PROC_EMPTY   ||
              proc_table[extra].status == PROC_ZOMBIE)
            {
              continue;
            }
          if (proc_table[pid].segment_i == proc_table[extra].segment_i)
            {
              proc_count++;
            }
        }
      if (proc_count == 0)
        {
          //
          // The code segment can be released, because no other
          // process, except the current one (to be closed),
          // is using it.
          //
          mb_free (proc_table[pid].address_i, proc_table[pid].size_i);
        }
    }
  //
  // Abandon children to `init' ((pid_t) 1).
  //
  for (child = 1; child < PROCESS_MAX; child++)
    {
      if (   proc_table[child].status != PROC_EMPTY
          && proc_table[child].ppid == pid)
        {
          proc_table[child].ppid = 1; // Son of `init'.
          if (proc_table[child].status == PROC_ZOMBIE)
            {
              sigchld = 1;    // Must send a SIGCHLD to `init'.
            }
        }
    }
  //
  // SIGCHLD to `init'.
  //
  if (   sigchld
      && pid != 1
      && proc_table[1].status != PROC_EMPTY
      && proc_table[1].status != PROC_ZOMBIE)
    {
      proc_sig_on ((pid_t) 1, SIGCHLD);
    }
  //
  // Announce to the parent the death of its child.
  //
  if (   pid != parent
      && proc_table[parent].status != PROC_EMPTY)
    {
      proc_sig_on (parent, SIGCHLD);
    }
}
