#include <kernel/proc.h>
#include <errno.h>
#include <kernel/k_libc.h>
//----------------------------------------------------------------------
static void sysroutine_error_back (int *number, int *line,
                                   char *file_name);
//----------------------------------------------------------------------
void
sysroutine (uint16_t *sp, segment_t *segment_d, uint16_t syscallnr,
            uint16_t msg_off, uint16_t msg_size)
{
  pid_t  pid      = proc_find (*segment_d);
  addr_t msg_addr = address (*segment_d, msg_off);
  //
  // Inbox.
  //
  union {
      sysmsg_chdir_t     chdir;
      sysmsg_chmod_t     chmod;
      sysmsg_chown_t     chown;
      sysmsg_clock_t     clock;
      sysmsg_close_t     close;
      sysmsg_dup_t       dup;
      sysmsg_dup2_t      dup2;
      sysmsg_exec_t      exec;
      sysmsg_exit_t      exit;
      sysmsg_fchmod_t    fchmod;
      sysmsg_fchown_t    fchown;
      sysmsg_fcntl_t     fcntl;
      sysmsg_fork_t      fork;
      sysmsg_fstat_t     fstat;
      sysmsg_kill_t      kill;
      sysmsg_link_t      link;
      sysmsg_lseek_t     lseek;
      sysmsg_mkdir_t     mkdir;
      sysmsg_mknod_t     mknod;
      sysmsg_mount_t     mount;
      sysmsg_open_t      open;
      sysmsg_read_t      read;
      sysmsg_seteuid_t   seteuid;
      sysmsg_setuid_t    setuid;
      sysmsg_signal_t    signal;
      sysmsg_sleep_t     sleep;
      sysmsg_stat_t      stat;
      sysmsg_stime_t     stime;
      sysmsg_time_t      time;
      sysmsg_uarea_t     uarea;
      sysmsg_umask_t     umask;
      sysmsg_umount_t    umount;
      sysmsg_unlink_t    unlink;
      sysmsg_wait_t      wait;
      sysmsg_write_t     write;
      sysmsg_zpchar_t    zpchar;
      sysmsg_zpstring_t  zpstring;
  } msg;
  //
  // Verify if the system call was emitted from kernel code.
  // The kernel can emit only some particular system call:
  //   SYS_NULL, to let other processes run;
  //   SYS_FORK, to let fork itself;
  //   SYS_EXEC, to replace a forked copy of itself.
  //
  if (pid == 0)
    {
      //
      // This is the kernel code!
      //
      if (   syscallnr != SYS_0
          && syscallnr != SYS_FORK
          && syscallnr != SYS_EXEC)
        {
          k_printf ("kernel panic: the system call %i ", syscallnr);
          k_printf ("was received while running in kernel space!\n");
        }
    }
  //
  // Entering a system call, the kernel variable `errno' must be
  // reset, otherwise, a previous kernel code error might be returned
  // to the applications.
  //
  errno    = 0;
  errln    = 0;
  errfn[0] = 0;
  //
  // Get message.
  //
  dev_io (pid, DEV_MEM, DEV_READ, msg_addr, &msg, msg_size, NULL);
  //
  // Do the request from the received system call.
  //
  switch (syscallnr)
    {
      case SYS_0:
          break;
      case SYS_CHDIR:
          msg.chdir.ret         = path_chdir (pid, msg.chdir.path);
          sysroutine_error_back (&msg.chdir.errno, &msg.chdir.errln,
                                 msg.chdir.errfn);
          break;
      case SYS_CHMOD:
          msg.chmod.ret         = path_chmod (pid, msg.chmod.path,
                                              msg.chmod.mode);
          sysroutine_error_back (&msg.chmod.errno, &msg.chmod.errln,
                                 msg.chmod.errfn);
          break;
      case SYS_CHOWN:
          msg.chown.ret         = path_chown (pid, msg.chown.path,
                                              msg.chown.uid,
                                              msg.chown.gid);
          sysroutine_error_back (&msg.chown.errno, &msg.chown.errln,
                                 msg.chown.errfn);
          break;
      case SYS_CLOCK:
          msg.clock.ret         = k_clock ();
          break;
      case SYS_CLOSE:
          msg.close.ret         = fd_close (pid, msg.close.fdn);
          sysroutine_error_back (&msg.close.errno, &msg.close.errln,
                                 msg.close.errfn);
          break;
      case SYS_DUP:
          msg.dup.ret           = fd_dup (pid, msg.dup.fdn_old, 0);
          sysroutine_error_back (&msg.dup.errno, &msg.dup.errln,
                                 msg.dup.errfn);
          break;
      case SYS_DUP2:
          msg.dup2.ret          = fd_dup2 (pid, msg.dup2.fdn_old,
                                           msg.dup2.fdn_new);
          sysroutine_error_back (&msg.dup2.errno, &msg.dup2.errln,
                                 msg.dup2.errfn);
          break;
      case SYS_EXEC:
          msg.exec.ret          = proc_sys_exec (sp, segment_d, pid,
                                                 msg.exec.path,
                                                 msg.exec.argc,
                                                 msg.exec.arg_data,
                                                 msg.exec.envc,
                                                 msg.exec.env_data);
          msg.exec.uid          = proc_table[pid].uid;
          msg.exec.euid         = proc_table[pid].euid;
          sysroutine_error_back (&msg.exec.errno, &msg.exec.errln,
                                 msg.exec.errfn);
          break;
      case SYS_EXIT:
          if (pid == 0)
            {
              k_printf ("kernel alert: "
                        "the kernel cannot exit!\n");
            }
          else
            {
              proc_sys_exit (pid, msg.exit.status);
            }
          break;
      case SYS_FCHMOD:
          msg.fchmod.ret        = fd_chmod (pid, msg.fchmod.fdn,
                                            msg.fchmod.mode);
          sysroutine_error_back (&msg.fchmod.errno, &msg.fchmod.errln,
                                 msg.fchmod.errfn);
          break;
      case SYS_FCHOWN:
          msg.fchown.ret        = fd_chown (pid, msg.fchown.fdn,
                                            msg.fchown.uid,
                                            msg.fchown.gid);
          sysroutine_error_back (&msg.fchown.errno, &msg.fchown.errln,
                                 msg.fchown.errfn);
          break;
      case SYS_FCNTL:
          msg.fcntl.ret         = fd_fcntl (pid, msg.fcntl.fdn,
                                            msg.fcntl.cmd,
                                            msg.fcntl.arg);
          sysroutine_error_back (&msg.fcntl.errno, &msg.fcntl.errln,
                                 msg.fcntl.errfn);
          break;
      case SYS_FORK:
          msg.fork.ret          = proc_sys_fork (pid, *sp);
          sysroutine_error_back (&msg.fork.errno, &msg.fork.errln,
                                 msg.fork.errfn);
          break;
      case SYS_FSTAT:
          msg.fstat.ret         = fd_stat (pid, msg.fstat.fdn,
                                           &msg.fstat.stat);
          sysroutine_error_back (&msg.fstat.errno, &msg.fstat.errln,
                                 msg.fstat.errfn);
          break;
      case SYS_KILL:
          msg.kill.ret          = proc_sys_kill (pid, msg.kill.pid,
                                                 msg.kill.signal);
          sysroutine_error_back (&msg.kill.errno, &msg.kill.errln,
                                 msg.kill.errfn);
          break;
      case SYS_LINK:
          msg.link.ret          = path_link (pid, msg.link.path_old,
                                             msg.link.path_new);
          sysroutine_error_back (&msg.link.errno, &msg.link.errln,
                                 msg.link.errfn);
          break;
      case SYS_LSEEK:
          msg.lseek.ret         = fd_lseek (pid, msg.lseek.fdn,
                                            msg.lseek.offset,
                                            msg.lseek.whence);
          sysroutine_error_back (&msg.lseek.errno, &msg.lseek.errln,
                                 msg.lseek.errfn);
          break;
      case SYS_MKDIR:
          msg.mkdir.ret         = path_mkdir (pid, msg.mkdir.path,
                                              msg.mkdir.mode);
          sysroutine_error_back (&msg.mkdir.errno, &msg.mkdir.errln,
                                 msg.mkdir.errfn);
          break;
      case SYS_MKNOD:
          msg.mknod.ret         = path_mknod (pid, msg.mknod.path,
                                              msg.mknod.mode,
                                              msg.mknod.device);
          sysroutine_error_back (&msg.mknod.errno, &msg.mknod.errln,
                                 msg.mknod.errfn);
          break;
      case SYS_MOUNT:
          msg.mount.ret         = path_mount (pid, msg.mount.path_dev,
                                              msg.mount.path_mnt,
                                              msg.mount.options);
          sysroutine_error_back (&msg.mount.errno, &msg.mount.errln,
                                 msg.mount.errfn);
          break;
      case SYS_OPEN:
          msg.open.ret          = fd_open (pid, msg.open.path,
                                           msg.open.flags,
                                           msg.open.mode);
          sysroutine_error_back (&msg.open.errno, &msg.open.errln,
                                 msg.open.errfn);
          break;
      case SYS_PGRP:
          proc_table[pid].pgrp          = pid;
          break;
      case SYS_READ:
          msg.read.ret          = fd_read (pid, msg.read.fdn,
                                           msg.read.buffer,
                                           msg.read.count,
                                           &msg.read.eof);
          sysroutine_error_back (&msg.read.errno, &msg.read.errln,
                                 msg.read.errfn);
          break;
      case SYS_SETEUID:
          msg.seteuid.ret       = proc_sys_seteuid (pid,
                                                    msg.seteuid.euid);
          msg.seteuid.euid      = proc_table[pid].euid;
          sysroutine_error_back (&msg.seteuid.errno, &msg.seteuid.errln,
                                 msg.seteuid.errfn);
          break;
      case SYS_SETUID:
          msg.setuid.ret        = proc_sys_setuid (pid,
                                                   msg.setuid.euid);
          msg.setuid.uid        = proc_table[pid].uid;
          msg.setuid.euid       = proc_table[pid].euid;
          msg.setuid.suid       = proc_table[pid].suid;
          sysroutine_error_back (&msg.setuid.errno, &msg.setuid.errln,
                                 msg.setuid.errfn);
          break;
      case SYS_SIGNAL:
          msg.signal.ret        = proc_sys_signal (pid,
                                                   msg.signal.signal,
                                                  msg.signal.handler);
          sysroutine_error_back (&msg.signal.errno, &msg.signal.errln,
                                 msg.signal.errfn);
          break;
      case SYS_SLEEP:
          proc_table[pid].status        = PROC_SLEEPING;
          proc_table[pid].ret           = 0;
          proc_table[pid].wakeup_events = msg.sleep.events;
          proc_table[pid].wakeup_signal = msg.sleep.signal;
          proc_table[pid].wakeup_timer  = msg.sleep.seconds;
          break;
      case SYS_STAT:
          msg.stat.ret          = path_stat (pid, msg.stat.path,
                                             &msg.stat.stat);
          sysroutine_error_back (&msg.stat.errno, &msg.stat.errln,
                                 msg.stat.errfn);
          break;
      case SYS_STIME:
          msg.stime.ret         = k_stime (&msg.stime.timer);
          break;
      case SYS_TIME:
          msg.time.ret          = k_time (NULL);
          break;
      case SYS_UAREA:
          msg.uarea.uid         = proc_table[pid].uid;
          msg.uarea.euid        = proc_table[pid].euid;
          msg.uarea.pid         = pid;
          msg.uarea.ppid        = proc_table[pid].ppid;
          msg.uarea.pgrp        = proc_table[pid].pgrp;
          msg.uarea.umask       = proc_table[pid].umask;
          strncpy (msg.uarea.path_cwd,
                   proc_table[pid].path_cwd, PATH_MAX);
          break;
      case SYS_UMASK:
          msg.umask.ret         = proc_table[pid].umask;
          proc_table[pid].umask = (msg.umask.umask & 00777);
          break;
      case SYS_UMOUNT:
          msg.umount.ret        = path_umount (pid,
                                               msg.umount.path_mnt);
          sysroutine_error_back (&msg.umount.errno, &msg.umount.errln,
                                 msg.umount.errfn);
          break;
      case SYS_UNLINK:
          msg.unlink.ret        = path_unlink (pid, msg.unlink.path);
          sysroutine_error_back (&msg.unlink.errno, &msg.unlink.errln,
                                 msg.unlink.errfn);
          break;
      case SYS_WAIT:
          msg.wait.ret          = proc_sys_wait (pid,
                                                 &msg.wait.status);
          sysroutine_error_back (&msg.wait.errno, &msg.wait.errln,
                                 msg.wait.errfn);
          break;
      case SYS_WRITE:
          msg.write.ret         = fd_write (pid, msg.write.fdn,
                                            msg.write.buffer,
                                            msg.write.count);
          sysroutine_error_back (&msg.write.errno, &msg.write.errln,
                                 msg.write.errfn);
          break;
      case SYS_ZPCHAR:
          dev_io (pid, DEV_TTY, DEV_WRITE, 0L, &msg.zpchar.c,
                  1, NULL);
          break;
      case SYS_ZPSTRING:
          dev_io (pid, DEV_TTY, DEV_WRITE, 0L,
                  msg.zpstring.string,
                  strlen (msg.zpstring.string), NULL);
          break;
      default:
          k_printf ("kernel alert: unknown system call %i!\n",
                    syscallnr);
          break;
    }
  //
  // Return value with a message back.
  //
  dev_io (pid, DEV_MEM, DEV_WRITE, msg_addr, &msg, msg_size, NULL);
  //
  // Continue with the scheduler.
  //
  proc_scheduler (sp, segment_d);
}
//----------------------------------------------------------------------
static void
sysroutine_error_back (int *number, int *line, char *file_name)
{
  *number = errno;
  *line   = errln;
  strncpy (file_name, errfn, PATH_MAX);
  file_name[PATH_MAX-1] = 0;
}
