#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
//----------------------------------------------------------------------
#define MEMORY_BLOCK_SIZE       1024
//----------------------------------------------------------------------
static char     _alloc_memory[LONG_BIT][MEMORY_BLOCK_SIZE];   // [1]
static size_t   _alloc_size[LONG_BIT];                        // [2]
static long int _alloc_map;                                   // [3]
//
// [1] Memory to be allocated.
// [2] Sizes allocated.
// [3] Memory block map. The memory map is made of a single integer and
//     the rightmost bit is the first memory block.
//----------------------------------------------------------------------
void *
malloc (size_t size)
{
    size_t   size_free;   // Size free found that might be allocated.
    int      m;           // Index inside `_alloc_memory[][]' table.
    int      s;           // Start index for a free memory area.
    long int mask;        // Mask to compare with `_alloc_map'.
    long int alloc;       // New allocation map.
    //
    // Check for arguments.
    //
    if (size == 0)
      {
        return (NULL);
      }
    //
    for (s = 0, m = 0; m < LONG_BIT; m++)
      {
        mask = 1;
        mask <<= m;
        //
        if (_alloc_map & mask)
          {
            //
            // The memory block is not free.
            //
            s         = m + 1;
            size_free = 0;
            alloc     = 0;
          }
        else
          {
            alloc     |= mask;
            size_free += MEMORY_BLOCK_SIZE;
          }
        if (size_free >= size)
          {
            //
            // Space found: update `_alloc_size[]' table, the map inside
            // `_alloc_map' and return the memory address.
            //
            _alloc_size[s] = size_free;
            _alloc_map    |= alloc;
            return ((void *) &_alloc_memory[s][0]);
          }
      }
    //
    // No space left.
    //
    errset (ENOMEM);                    // Not enough space.
    //
    return (NULL);
}
//----------------------------------------------------------------------
void
free (void *address)
{
    size_t   size_free;  // Size to make free.
    int      m;          // Index inside `_alloc_memory[][]' table.
    int      s;          // Start index.
    long int mask;       // Mask to compare with `_alloc_map'.
    long int alloc;      // New allocation map.
    //
    // Check argument.
    //
    if (address == NULL)
      {
        return;
      }
    //
    // Find the original allocated address inside `_alloc_memory[][]'
    // table.
    //
    for (m = 0; m < LONG_BIT; m++)
      {
        if (address == (void *) &_alloc_memory[m][0])
          {
            //
            // This is the right memory block.
            //
            if (_alloc_size[m] == 0)
              {
                //
                // The block found is not allocated.
                //
                return;
              }
            else
              {
                //
                // Build the map of the memory to set free.
                //
                size_free = _alloc_size[m];
                for (alloc = 0, s = m;
                     size_free > 0 && s < LONG_BIT;
                     size_free -= MEMORY_BLOCK_SIZE, s++)
                  {
                    mask   = 1;
                    mask <<= s;
                    alloc |= mask;
                  }
                //
                // Compare the map of memory to be freed with the
                // reality allocated one, then free the memory.
                //
                if ((_alloc_map & alloc) == alloc)
                  {
                    _alloc_map    &= ~alloc;
                    _alloc_size[m] = 0;
                    return;
                  }
                //
                // The real map does not report the same amount of
                // allocated memory, so nothing is freed.
                //
                return;
              }
          }
      }
    //
    // Address not allocated.
    //
    return;
}
//----------------------------------------------------------------------
void *
realloc (void *address, size_t size)
{
    char  *address_new;
    char  *address_old = (char *) address;
    size_t size_old    = 0;
    size_t size_new    = size;
    int    m;                   // Index inside the memory table;
    //
    // Check arguments.
    //
    if (size == 0)              return (NULL);
    if (address == NULL)        return (malloc (size));
    //
    // Locate original allocation.
    //
    for (m = 0; m < LONG_BIT; m++)
      {
        if (address_old == (char *) &_alloc_memory[m][0])
          {
            size_old = _alloc_size[m];
            break;
          }
      }
    //
    // Check if a valid size was found.
    //
    if (size_old == 0)
      {
        //
        // Address not found or size not valid.
        //
        return (NULL);
      }
    //
    // Allocate the new memory.
    //
    address_new = malloc (size);
    //
    // Check allocation. If there is an error, the variable `errno'
    // is already updated by `malloc()'.
    //
    if (address_new == NULL)
      {
        return (NULL);
      }
    //
    // Copy old memory.
    //
    for (; size_old > 0 && size_new > 0;
         size_old--, size_new--, address_new++, address_old++)
      {
       *address_new = *address_old;
      }
    //
    // Free old memory.
    //
    free (address);
    //
    // Return the new address.
    //
    return (address_new);
}
//----------------------------------------------------------------------
